{Notes a propos de la rsolution :

Plusieurs techniques de rsolution utilisent le meme type d'algo. C'est le cas pour
CandidatsNiemes, NiemesCachs, EtoileANBranches.

Les explications suivantes sont pour CandidatNiemes :

Pour ces techniques, j'utilise un algo rcursif qui recoit un lment
(= NoElement ET TZone prdfini).                                      //1\\
Pour cet lment, on passe plusieurs indexs et groupe d'indexs         //2\\
(c'est a dire que a chaque rcursion, on prend une valeur en plus).
Pour chaque indexs, on prends les candidats                            //3\\
et on en fait la SOMME D'ENSEMBLES.

Exemple
  Index     :        1          3           9
  Candidats :     [1,5,6]    [2,5,8]    [1,2,6,8]

  SommeCand : [1,2,5,6,8]

Ensuite, des calculs sont faits pour savoir si
 - le nombre de Indexs (ici 3) est gal au nombre de Candidats dans SommeCand (ici 5)   (*).

Pour amliorer ces algos, des tests sont souvent faits sur les
indexs, (avant d'en utiliser les Candidats) pour savoir si l'index permet de vrifier (*)


Pour les NiemesCachs, EtoileANBranches, la technique est la meme. Ce tableau permet de voir
ce qui change :

--------------------------------------------------------------------------------
Mthode de            |   CandidatsNieme  |   NiemeCach  |  EtoileANBranches  |
rsolution            |                   |               |                    |
--------------------------------------------------------------------------------
Parametres invariants |   NoElement et    |  NoElement et |     Valeur         |
= //1\\               |     TZone         |    TZone      |   et TZone         |
--------------------------------------------------------------------------------
Variable  utiliser   |     Indexs        |   Valeurs     |  NoElement         |
pour la rcursion     |                   |               |                    |
= //2\\               |                   |               |                    |
--------------------------------------------------------------------------------
Variable de calcul    |   Valeurs         |   Indexs      |     Indexs         |
dans les ensembles    | (= Candidats ici) |               |                    |
= //3\\               |                   |               |                    |
--------------------------------------------------------------------------------


////////////////////////////////////
//      Un exemple complet :      //
////////////////////////////////////

Je cherche un candidat triple (= dans 3 cases, il existe au plus 3 valeurs diffrentes)
voici mon lment :

                            **      **            **
|  1,3  |  2  | 1,4,6,9 |  4,5  |  4,8  |  7  | 4,5,8 |  1,9  |  3,4,8 |

On voit un candidat triple de 4,5,8 dans les cases (**)

D'apres le tableau, nous allons nous servir des indexs pour la rcursion.

On prends l'Index 1 : 1,3
On a 1 index et 2 candidats (1<>2 donc on ajoute un index)

  On prends l'index 2 : 2
  Il est marqu comme trouv, on change cet index

  On prends l'index 3 : 1,4,6,9
  On fait la somme des candidats : 1,3,4,6,9
  On a 2 indexs et 5 candidats (on a interdit a l'algo d'aller a plus de 4 candidats, donc on annule cet index)

  .....

  On a test l'index 9 qui ne fonctionne pas non plus, donc on reviens a la procdure d'avant


On remarque qu'avec l'index 1, on ne trouve rien
On remarque qu'avec l'index 2, on ne trouve rien
On remarque qu'avec l'index 3, on ne trouve rien

On prends l'Index 4 : 4,5
On a 1 index et 2 candidats (1<>2 donc on continue)

  On prends l'Index 5 : 4,8  (pas la peine d'aller voir dans les indexs prcdents)
  On fait la somme des candidats : 4,5,8
  On a 2 indexs et 3 candidats (2<>3 donc on ajoute un index)

    On prends l'index 6 : 7
    Il est marqu comme trouv, on change cet index

    On prends l'index 7 : 4,5,8
    On fait la somme des candidats : 4,5,8
    On a 3 indexs et 3 candidats .... BINGO

    A ce stade, il est parfois possible qu'il y ait un test supplmentaire, mais
    avec CandidatsNieme, pas besoin. Si ce test n'est pas concluant, on change l'index

    On supprime les candidats appropris. Si aucun n'est supprim, on change l'index
    Si il y a eu suppression, on quitte la procedure


}
unit USudokuResolution;

interface

uses
  Classes, SysUtils, Dialogs;


const TAILLE_SUDOKU = 9;
      TAILLE_BLOC = 3;
      POS_BLOC = 7;             // 2 puissance TAILE_BLOC - 1 = 0000 0000 0000 0111
                                // utile pour PositionDesIndexs
      POS_BLOC2 = 73;            // 0000 0000 0100 1001 = TAILLE_BLOC "1" spars de TAILLE_BLOC-1 "0"
                                //                       (trois "1" spars de 2 "0")

      ENSEMBLE_VALEURS = $01FF;           //0000 0001 1111 1111

type
  TSmallNum = 0..TAILLE_SUDOKU;                //type utilis a la place des integer (gain de place ?)
  TEnsemble = Word;                            //type d'un candidat par exemple
  TZone = (zsLigne, zsColonne, zsBloc);        //type de zone a traiter

  TResultat = record                              //Type de retour des fonctions
    Index      : TSmallNum;
    NoElements : TEnsemble;
    Zone       : TZone;
  end;


  TCaseSudoku = Record
     Cand   : TEnsemble;
     Trouve : Boolean;
  end;

  TIndexs = Array of TSmallNum;

  TGrilleSudoku = array [1..TAILLE_SUDOKU,1..TAILLE_SUDOKU] of TCaseSudoku;


  PCasesSudoku = array [1..TAILLE_SUDOKU] of ^TCaseSudoku;
    //Pointeurs sur une zone de la grille
  PZonesSudoku = array [TZone,1..TAILLE_SUDOKU] of PCasesSudoku;
    //Forme [TZone,NoElement,Index]
    //Pointeurs sur toutes les zones de la grille (chaque case est pointe 3 fois ...)
    //Ce type permet de passer en parametre a une fonction une partie de la grille


  TZoneOccurence = array [1..TAILLE_SUDOKU] of TSmallNum;
    //Permet de donner a la fonction les occurences d'une seule zone
  TOccurence = array [TZone,1..TAILLE_SUDOKU] of TZoneOccurence;
    //Occurence contient le nb de V dans la zone numro N
    //   Exemple : Occurence[zsLigne,2,6] contient le nombre de 6 dans la ligne 2)
    // les chiffres dja placs ne comptent pas

var
    AfficherEtapesResolution : Boolean;
    EtapesResolution : String;



  procedure IndexToPosition (const Index, NoElement : TSmallNum; const Zone : TZone; out Ligne, Colonne : TSmallNum);

  procedure Initialiser(var Grille : TGrilleSudoku; var Occur : TOccurence);

  function EnregistrerTableau (Ligne : String; var Grille : PZonesSudoku; var Occur : TOccurence):Byte;

  function CreerPointeursZonesSudoku (var Grille : TGrilleSudoku) : PZonesSudoku;

  function TesterCandidatSeul (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum) : TResultat;

  function TesterEmplacementUnique (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum) : TResultat;

  function TesterCandidatsJusquaNieme (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum):TResultat;

  function TesterInteraction (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum) : TResultat;

  function TesterEtoileANBranche (var Grille : PZonesSudoku; var Occ : TOccurence; Var LastValeur : TSmallNum): TResultat;

  function ResoudreBruteForce (Sud : TGrilleSudoku; NoCase : Byte; LastValeur : TSmallNum; var GrilleSortie : TGrilleSudoku):Boolean;

  function ResoudreBruteForce2 (var Sud : TGrilleSudoku; NoCase : Byte):boolean;

  function ResoudreBruteForce3 (var Ptrs : PZonesSudoku; NoCase : Byte):boolean;


  function AjouterValeurBF3 (var Grille : PZonesSudoku; Valeur, Index, NoElement : TSmallNum; Zone : TZone):Boolean;




implementation




uses UManipBits;


const ZoneToStr : array [TZone] of String = (' la ligne ', ' la colonne ', ' le bloc ');
      ZonePluToStr : array [TZone] of String = (' les lignes ', ' les colonnes ', ' les blocs ');





{------------------------------------------------------------}
{                                                            }
{                     Methodes generales                     }
{                                                            }
{------------------------------------------------------------}



{function Modulo (const N : Shortint; Facteur : TSmallNum) : TSmallNum;
//Modulo plus rapide que mod
begin
  Result:=N;
  While Result>=Facteur do
    Dec(Result,Facteur);
end;                              }





function LCToBloc (Ligne, Colonne : TSmallNum) :TSmallNum;
//Recupere le numro du bloc associ a la ligne et colonne
//Les blocs sont numrots :
//  1 2 3
//  4 5 6
//  7 8 9

begin
  Result:=((Ligne-1) div TAILLE_BLOC)*TAILLE_BLOC + ((Colonne-1) div TAILLE_BLOC) + 1;
end;




procedure IndexToPosition (const Index, NoElement : TSmallNum; const Zone : TZone; out Ligne, Colonne : TSmallNum);
//Transforme des coordonnes Index, NoElement en Ligne, Colonne
begin
  case Zone of
    zsLigne   : begin
                  Ligne:=NoElement;
                  Colonne:=Index;
                end;
    zsColonne : begin
                  Ligne:=Index;
                  Colonne:=NoElement;
                end;
    zsBloc    : begin
                  Ligne:=((NoElement-1) div TAILLE_BLOC)*TAILLE_BLOC+(Index-1) div TAILLE_BLOC+1;
                  Colonne:=((NoElement-1) mod TAILLE_BLOC)*TAILLE_BLOC+(Index-1) mod TAILLE_BLOC+1;
                end;
  end;
end;





function EnsembleToStr (Ens : TEnsemble): string;
var Bit : TSmallNum;
begin
  Result:='';
  for Bit:=0 to TAILLE_SUDOKU-1 do
    if IsBitSet(Ens,Bit) then Result:=Result+IntToStr(Bit+1)+', ';

  Delete(Result,Length(Result)-1,2);
end;





{------------------------------------------------------------}
{                                                            }
{            Methodes lies a l'ajout de valeurs             }
{                                                            }
{------------------------------------------------------------}



function EnleverCandidat(var CaseCand : TEnsemble; var Occur : TOccurence; Valeur, Index, NoElement : TSmallNum; Zone : TZone) : Boolean;
//Enleve le candidat 'Valeur' de la case 'Ligne, Colonne' de CaseCand
//Met galement a jour les valeurs du tableau Occurence (enleve 1 dans la ligne colonne et bloc si le chiffre est prsent dans CaseCand)
//Procedure appele uniquement dans 'SupprimerCandidatsZone' et 'AjouterValeur'
var Ligne, Colonne : TSmallNum;
begin
  Result:=False;
  if IsBitSet(CaseCand,Valeur-1) then
  begin
    IndexToPosition(Index, NoElement, Zone, Ligne, Colonne);
    ClearBit(CaseCand,Valeur-1);
    dec(Occur[zsLigne,Ligne,Valeur]);
    dec(Occur[zsColonne,Colonne,Valeur]);
    dec(Occur[zsBloc,LCToBloc(Ligne, Colonne),Valeur]);
    Result:=True;
  end;
end;





procedure SupprimerCandidatsZone(var ZoneGrille : PCasesSudoku; var Occur : TOccurence; Valeur, NoElement : TSmallNum; const Zone : TZone);
//Enleve tout les candidats V d'une zone prcise
//Le type de zone est spcifi par 'Zone'
//Le numro de la zone est 'NoElement' et va de 1  9

var Index : TSmallNum;
begin
  for Index:=1 to TAILLE_SUDOKU do
    if not(ZoneGrille[Index]^.Trouve) then EnleverCandidat(ZoneGrille[Index]^.Cand, Occur, Valeur, Index, NoElement, Zone);
end;





procedure AjouterValeur (var Grille : PZonesSudoku; var Occur : TOccurence; Valeur, Index, NoElement : TSmallNum; Zone : TZone);
//Procedure qui ajoute une valeur 'Valeur' en position 'Ligne, Colonne'
//Elle ajoute la valeur dans Grille, enleve les candidats appropris
//et tient a jour le tableau 'Occurence'

var V, NoLigne, NoColonne, NoBloc : TSmallNum;
begin
  IndexToPosition(Index, NoElement, Zone, NoLigne, NoColonne);
  NoBloc:=LCToBloc(NoLigne,NoColonne);

  SupprimerCandidatsZone (Grille[zsLigne,NoLigne],     Occur, Valeur, NoLigne, zsLigne);
  SupprimerCandidatsZone (Grille[zsColonne,NoColonne], Occur, Valeur, NoColonne, zsColonne);
  SupprimerCandidatsZone (Grille[zsBloc,NoBloc],       Occur, Valeur, NoBloc, zsBloc);
    //Supprime les candidats dans la meme ligne, meme colonne et meme bloc


  for V:=1 to TAILLE_SUDOKU do
  if IsBitSet(Grille[Zone,NoElement,Index]^.Cand,V-1) then
    EnleverCandidat(Grille[Zone,NoElement,Index]^.Cand, Occur, V, Index, NoElement, Zone);

  SetBit(Grille[Zone,NoElement,Index]^.Cand,Valeur-1);
  Grille[Zone,NoElement,Index]^.Trouve:=True;
    //Supprime les candidats dans la case ou on va ajouter la valeur
end;





{------------------------------------------------------------}
{                                                            }
{                 Methodes d'initialisation                  }
{                                                            }
{------------------------------------------------------------}



function CreerPointeursZonesSudoku (var Grille : TGrilleSudoku) : PZonesSudoku;
//Cree un tableau contenant des pointeurs sur toutes les cases du tableau.
//Permet de facilement modifier une case en connaissant son NoElement, et son Index
var NoElement,Index : TSmallNum;
begin
  for NoElement:=1 to TAILLE_SUDOKU do
    for Index := 1 to TAILLE_SUDOKU do
    begin
      Result[zsLigne,NoElement,Index]:=@Grille[NoElement,Index];
      Result[zsColonne,Index,NoElement]:=@Grille[NoElement,Index];
      Result[zsBloc,((NoElement-1) div TAILLE_BLOC)*TAILLE_BLOC+(Index-1) div TAILLE_BLOC+1,
            ((NoElement-1) mod TAILLE_BLOC)*TAILLE_BLOC+(Index-1) mod TAILLE_BLOC +1]:=@Grille[NoElement,Index];
    end;
end;




procedure Initialiser(var Grille : TGrilleSudoku; var Occur : TOccurence);
//Initialise le tableau Grille et Occurence

var i,j : TSmallNum;
begin
  for i:=1 to TAILLE_SUDOKU do
    for j:=1 to TAILLE_SUDOKU do
    begin
      Grille[i,j].Trouve:=False;
      Grille[i,j].Cand:=ENSEMBLE_VALEURS;
      Occur[zsLigne,i,j]:=TAILLE_SUDOKU;
      Occur[zsColonne,i,j]:=TAILLE_SUDOKU;
      Occur[zsBloc,i,j]:=TAILLE_SUDOKU;
    end;
end;





function EnregistrerTableau (Ligne : String; var Grille : PZonesSudoku; var Occur : TOccurence):Byte;
//Retourne les valeurs du fichier
//Ne pas oublier d'initialiser les tableaux avant !

var i,j : TSmallNum;
    C : Char;
begin
  Result:=TAILLE_SUDOKU*TAILLE_SUDOKU;
  for i:=1 to TAILLE_SUDOKU do
    for j:=1 to TAILLE_SUDOKU do
    begin
      C:=Ligne[(i-1)*TAILLE_SUDOKU+j];
      if C in ['1'..'9'] then
      begin
        AjouterValeur(Grille, Occur, StrToInt(C), j, i, zsLigne);  //Cette ligne est un peu bizarre. En fait, la configuration Ligne/Colonne correspond
                                                                   //a la config NoElement/Index d'un zsLigne. Autrement dit, pour parcourir la grille
                                                                   //par ligne et colonne, il suffit de la parcourir en NoElement/Index avec une zone de type zsLigne
        dec(Result);
      end;
    end;
end;





{------------------------------------------------------------}
{                                                            }
{                   Methodes de rsolution                   }
{                                                            }
{------------------------------------------------------------}





function CompteCandidats(C : TEnsemble) :TSmallNum;
//Renvoie le nombre de candidats de l'argument 'C'

var V : TSmallNum;
begin
  Result := 0;
  for V:=0 to TAILLE_SUDOKU-1 do
    Inc(Result,(C shr V) and 1);
end;





function DerniereValeurInEns(Ens : TEnsemble) :TSmallNum;
//Renvoie le dernier candidat prsent dans C
//Par exemple, [1,5,6] -> 6, [7] -> 7, [] -> 0
//Sert seulement pour le "CandidatSeul"

begin
  Result:=TAILLE_SUDOKU;
  while (Result>0) and not(IsBitSet(Ens,Result-1)) do
    dec(Result);
end;





procedure MessageCandidatSeul(Valeur :TSmallNum; C : TResultat);
begin
 if EtapesResolution <> '' then
 EtapesResolution:=EtapesResolution+#13#10;
 EtapesResolution:=EtapesResolution+' - Ajout d''un '+IntToStr(Valeur)+' en '
  +EnsembleToStr(C.NoElements)+', '+IntToStr(C.Index)+' car candidat unique.';
end;






function TesterCandidatSeul (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum) : TResultat;
//Renvoie la case (type TCase) de la premiere cellule du tableau (a partir de LastRow)
//ne contenat qu'un seul candidat. Ajoute aussi une ligne au mmo.
//Renvoie 0,0,0 si aucune cellule du tableau n'a qu'un seul candidat
//Ajoute aussi la valeur dans la Grille
//LastElement appartient [1..9]

var NoElement, Index, ElemTemp : TSmallNum;
    CaseTemp : TCaseSudoku;
begin
  Result.NoElements:=0;
  Result.Index     :=0;


  for NoElement:=1 to TAILLE_SUDOKU do
    for Index:=1 to TAILLE_SUDOKU do
    begin
      ElemTemp:=((NoElement+LastElement-2) mod TAILLE_SUDOKU)+1;
      CaseTemp:=Grille[zsLigne,ElemTemp,Index]^;

      if not(CaseTemp.Trouve) and (CompteCandidats(CaseTemp.Cand)=1) then
      begin
        Result.NoElements:=1 shl (ElemTemp-1);
        Result.Index:=Index;
        Result.Zone:=zsLigne;

        AjouterValeur(Grille, Occur, DerniereValeurInEns(CaseTemp.Cand), Index, ElemTemp, zsLigne);
        LastElement:=ElemTemp;
        if AfficherEtapesResolution then MessageCandidatSeul(DerniereValeurInEns(CaseTemp.Cand),Result);
        Exit;  //<-- Permet pas mal de lignes en moins et de la lisibilit en plus !!
      end;
    end;
end;




function ChercherIndexInZone(var ZoneGrille : PCasesSudoku; Valeur : TSmallNum) : TSmallNum;
//Recherche dans une zone du tableau la premiere cellule contenant 'Valeur' (pas forcment seul)
//Donner a la fonction un tableau de la taille TAILLE_SUDOKU
//La zone est dfinie par 'Zone' et son numro par 'NoElement'
//Renvoie 0 si 'Valeur' n'existe pas dans la zone spcifie, renvoie l'Index sinon

var Index : TSmallNum;
begin
  Result:=0;
  for Index:=1 to TAILLE_SUDOKU do
    if not(ZoneGrille[Index]^.Trouve) and IsBitSet(ZoneGrille[Index]^.Cand,Valeur-1) then
    begin
      Result:=Index;
      Exit;
    end;
end;




function CompterCasesNonTrouveInZone(var ZoneGrille : PCasesSudoku) : TSmallNum;
//Compte le nombre de cases non trouvees dans une zone
var Index : TSmallNum;
begin
  Result:=0;
  for Index:=1 to TAILLE_SUDOKU do
    if not(ZoneGrille[Index]^.Trouve) then Inc(Result);
end;





procedure MessageEmplacementUnique (Valeur : TSmallNum; R : TResultat);
//Affiche le message appropri pour un lment seul
var Ligne, Colonne, NoElem : TSmallNum;
begin
  if EtapesResolution <> '' then
  EtapesResolution:=EtapesResolution+#13#10;
  NoElem:=DerniereValeurInEns(R.NoElements);
  IndexToPosition(R.Index,NoElem,R.Zone,Ligne,Colonne);
  EtapesResolution:=EtapesResolution+' - Ajout d''un '+IntToStr(Valeur)+' en '+IntToStr(Ligne)+', '+IntToStr(Colonne)
                                           +' car valeur unique dans'+ZoneToStr[R.Zone]+IntToStr(NoElem);
end;





function EmplacementUniqueInZone(var Grille : PZonesSudoku; var Occur : TOccurence; NoElement : TSmallNum; Zone : TZone) : TResultat;
//Cherche si dans la zone spcifie si il existe une valeur ayant un seul emplacement possible
//si oui, renvoie un TResultat
var Valeur, Index : TSmallNum;
begin
  Result.Index:=0;
  Result.NoElements:=0;

  for Valeur:=1 to TAILLE_SUDOKU do
    if Occur[Zone,NoElement,Valeur]=1 then
    begin
      Index:=ChercherIndexInZone(Grille[Zone,NoElement],Valeur);
      Result.Index:=Index;
      Result.NoElements:=1 shl (NoElement-1);
      Result.Zone:=Zone;
      AjouterValeur(Grille,Occur,Valeur,Index,NoElement,Zone);
      if AfficherEtapesResolution then MessageEmplacementUnique(Valeur,Result);
      Exit;
    end;
end;






function TesterEmplacementUnique (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum) : TResultat;
//Cherche dans chaque ligne, chaque colonne et chaque bloc si une valeur apparait une seule fois
//Cette fonction s'arrete au premier succs (une valeur n'apparait qu'une fois)
//Affiche une ligne au mmo 'Etapes'
//Renvoie 0,0,0 si aucune valeur n'est seule dans n'importe quelle zone

var NoElement, NoElemTemp, OrdZone : TSmallNum;
    ValTemp : TResultat;
begin
  Result.NoElements:=0;
  Result.Index:=0;

  for NoElement:=1 to TAILLE_SUDOKU do
  begin
    NoElemTemp:=((NoElement+LastElement-2) mod TAILLE_SUDOKU)+1;
    for OrdZone:=0 to 2 do
    begin
      ValTemp:=EmplacementUniqueInZone(Grille,Occur,NoElemTemp,TZone(OrdZone));
      if ValTemp.Index<>0 then
      begin
        Result:=ValTemp;
        Exit;
      end;
    end;
  end;

end;






function SupprimerDansIndexsInZone (var ZoneGrille : PCasesSudoku; var Occur : TOccurence; Cand, Indexs : TEnsemble; NoElement : TSmallNum; Zone : TZone) : Boolean;
//supprime les candidats de Cand dans les cases repres par Indexs
var Index, Valeur : TSmallNum;
            Modif : boolean;
begin
  Result:=False;
  for Index:=0 to TAILLE_SUDOKU-1 do
  if IsBitSet(Indexs,Index) then
    for Valeur:=0 to TAILLE_SUDOKU-1 do
    if IsBitSet(Cand,Valeur) then
    begin
      Modif:=EnleverCandidat(ZoneGrille[Index+1]^.Cand,Occur,Valeur+1,Index+1,NoElement,Zone);
      Result:=Result or Modif;
    end;
end;





procedure MessageCandidatNieme(NbCand : TSmallNum; Cand, Indexs : TEnsemble; NoElement : TSmallNum; Zone : TZone);
const TypeToStr : array [2..4] of string = ('doubles','triples','quadruples');
var Positions, Cands : String;
    Valeur, Index, Ligne, Colonne : TSmallNum;
begin
  Cands:='';
  for Valeur:=0 to TAILLE_SUDOKU-1 do
  if IsBitSet(Cand,Valeur) then
    Cands:=Cands+IntToStr(Valeur+1)+', ';
  Delete(Cands,Length(Cands)-1,2);

  Positions:='';
  for Index:=0 to TAILLE_SUDOKU-1 do
  if IsBitSet(Indexs,Index) then
  begin
    IndexToPosition(Index+1,NoElement,Zone,Ligne,Colonne);
    Positions:=Positions+'('+IntToStr(Ligne)+','+IntToStr(Colonne)+'), ';
  end;
  Delete(Positions,Length(Positions)-1,2);

  if EtapesResolution <> '' then
  EtapesResolution:=EtapesResolution+#13#10;

  EtapesResolution:=EtapesResolution+' - Suppression de candidats dans'+ZoneToStr[Zone]+IntToStr(NoElement)+' car candidats '+TypeToStr[NbCand]
                                           +' de '+Cands+' dans les cases '+Positions;
end;





function TrouverCandidatJusquaNieme (var ZoneGrille : PCasesSudoku; var Occur : TOccurence; NoElement : TSmallNum; Zone : TZone;
    {Parametres pour la recursion}   NMax, NCases, IndexPrec : TSmallNum; Somme, Indexs : TEnsemble) : boolean;

var Index, NbCand : TSmallNum;
    SommeInter, IndexsInter : TEnsemble;
begin
  Result:=False;
  for Index:=IndexPrec to TAILLE_SUDOKU do
  if not(ZoneGrille[Index]^.Trouve) then
  begin
    SommeInter:=Somme or ZoneGrille[Index]^.Cand;
    NbCand:=CompteCandidats(SommeInter);
    if (NbCand<=NMax) then
    begin
      IndexsInter:=Indexs or (1 shl (Index-1));
      if (NCases<NbCand) and (Index<TAILLE_SUDOKU)  then Result:=TrouverCandidatJusquaNieme(ZoneGrille,Occur,NoElement,Zone,NMax,NCases+1,Index+1,SommeInter,IndexsInter)
      else
        if NCases=NbCand then
        begin
          Result:=SupprimerDansIndexsInZone(ZoneGrille,Occur,SommeInter,ENSEMBLE_VALEURS-IndexsInter,NoElement,Zone);
          if AfficherEtapesResolution and Result then MessageCandidatNieme(NCases,SommeInter,IndexsInter,NoElement,Zone);
        end;

      if Result then Break;
    end;
  end;
end;



{---------------------------------------------------------------------------------------------
ICI c'est une vieille fonction qui faisait la meme chose que celle du dessus, mais celle la
permettait de trouver par exemple que des candidats triples, alors que la premiere trouve
n'importe quel "paquet" de taille < NMax    (paquet = candidats doubles, triples, etc ...)



function TrouverCandidatNieme(var TabIndex : array of TSmallNum; const N, NRestant, LastIndex : TSmallNum; const Somme : TCandidats;
            NoElement : TSmallNum; Zone : TZone) : Boolean;
//Recursif !
//Au dpart, spcifier a la fonction un TabIndex vide, de taille plus grande que N.
//N correspond au type de recherche : Candidats doubles (N=2), triples (N=3) jusqu'a, en thorie, Octuples (N=8)
//NRestant correspond au nombre d'index restants, sans compter la fonction en cours (mettre = a N-1 au dpart)
//Spcifier LastIndex a 0 pour le dpart. Correspond en fait a l'index prcdent, c'est a partir de lui qu'il faut chercher
//Somme : correspond a la somme de tous les candidats des index prcdents
//TabIndex : contient (a la fin) les indexs correspondants au CandidatNieme dans l'ordre dcroissant

//pour trouver si un index est candidatNieme, on fait la somme des candidats des index prcdents. si le nombre de valeurs
//de cette somme est plus grand que N, l'index ne fait pas partie du candidat Nieme des index prcdents

var Index : TSmallNum;
    Ligne, Colonne : TSmallNum;
    SommeInter : TCandidats;
begin
  Result:=False;
  Index:=LastIndex;                     //Commencer la recherche a partir de l'index prcdent +1 ( attention : LastIndex = [1..9] mais Index = [0..8]
  while (Index<9-NRestant) and not Result do
  begin
    IndexToPosition(Index+1, NoElement, Zone, Ligne, Colonne);

    if Candidats[Ligne,Colonne]<>[] then                 //Verifie simplement qu'il y a des candidats
    begin
      SommeInter:=Somme+Candidats[Ligne, Colonne];
      if CompteCandidats(SommeInter)<=N then
      begin
        TabIndex[NRestant]:=Index+1;
        if NRestant=0 then
        begin
          Result:=SupprimerSaufTIndexInZone(SommeInter,TabIndex,NoElement,Zone);
          if Result and AffMess then MessageCandidatNieme(SommeInter, TabIndex, N, NoElement, Zone);
        end
        else
          Result:=TrouverCandidatNieme(TabIndex,N,NRestant-1,Index+1,SommeInter,NoElement,Zone);
      end;
    end;

    inc(Index);
  end;
end;

---------------------------------------------------------------------------------------------}

function GetIndexs (var ZoneGrille : PCasesSudoku; Valeur : TSmallNum) : TEnsemble;
var Index : TSmallNum;
begin
  Result:=0;
  for Index:=1 to TAILLE_SUDOKU do
    if IsBitSet(ZoneGrille[Index]^.Cand,Valeur-1) then SetBit(Result,Index-1);
end;







procedure MessageNiemeCache(NbCand : TSmallNum; Cand, Indexs : TEnsemble; NoElement : TSmallNum; Zone : TZone);
const TypeToStr : array [2..4] of string = ('paire cache','triplet cach','quadruplet cach');
var Positions, Cands : String;
    Valeur, Index, Ligne, Colonne : TSmallNum;
begin
  Cands:='';
  for Valeur:=0 to TAILLE_SUDOKU-1 do
  if IsBitSet(Cand,Valeur) then
    Cands:=Cands+IntToStr(Valeur+1)+', ';
  Delete(Cands,Length(Cands)-1,2);

  Positions:='';
  for Index:=0 to TAILLE_SUDOKU-1 do
  if IsBitSet(Indexs,Index) then
  begin
    IndexToPosition(Index+1,NoElement,Zone,Ligne,Colonne);
    Positions:=Positions+'('+IntToStr(Ligne)+','+IntToStr(Colonne)+'), ';
  end;
  Delete(Positions,Length(Positions)-1,2);

  if EtapesResolution <> '' then
  EtapesResolution:=EtapesResolution+#13#10;

  EtapesResolution:=EtapesResolution+' - Suppression de candidats dans'+ZoneToStr[Zone]+IntToStr(NoElement)+' car '+TypeToStr[NbCand]
                                           +' de '+Cands+' dans les cases '+Positions;
end;





function TrouverCacheJusquaNieme (var ZoneGrille : PCasesSudoku; var Occur : TOccurence; NoElement : TSmallNum; Zone : TZone;
    {Parametres pour la recursion}   NMax, NbValeurs, ValeurPrec : TSmallNum; SommeIndexs, Valeurs : TEnsemble) : boolean;

var Valeur, NbIndexs : TSmallNum;
    ValeursTemp, SommeIndexsTemp : TEnsemble;
begin
  Result:=False;
  for Valeur:=ValeurPrec to TAILLE_SUDOKU do
  if (Occur[Zone,NoElement,Valeur]>=2) and (Occur[Zone,NoElement,Valeur]<=NMax) then
  begin
    SommeIndexsTemp:=SommeIndexs or GetIndexs(ZoneGrille,Valeur);
    NbIndexs:=CompteCandidats(SommeIndexsTemp);
    if (NbIndexs>=NbValeurs) then
    begin
      ValeursTemp:=Valeurs or (1 shl (Valeur-1));
      if (NbValeurs<NMax) and (NbValeurs<NbIndexs) and (Valeur<TAILLE_SUDOKU) then
        Result:=TrouverCacheJusquaNieme(ZoneGrille,Occur,NoElement,Zone,NMax,NbValeurs+1,Valeur+1,SommeIndexsTemp,ValeursTemp)
      else
        if NbValeurs=NbIndexs then
        begin
          Result:=SupprimerDansIndexsInZone(ZoneGrille,Occur,ENSEMBLE_VALEURS-ValeursTemp,SommeIndexsTemp,NoElement,Zone);
          if AfficherEtapesResolution and Result then MessageNiemeCache(NbValeurs,ValeursTemp,SommeIndexsTemp,NoElement,Zone);
        end;

      if Result then Break;
    end;
  end;
end;




{---------------------------------------------------------------------------------------------


function TrouverNiemeCache(var TabValeurs : array of TSmallNum; const N, NRestant, LastValeur : TSmallNum; const Indexs : TCandidats;
            NoElement : TSmallNum; Zone : TZone) : Boolean;
//Recursif !
//Au dpart, spcifier a la fonction un TabValeurs vide, de taille egale a N.
//N correspond au type de recherche : paire cahce (N=2), triplets cachs (N=3) jusqu'a, en thorie, Octuplets (N=8)
//NRestant correspond au nombre d'index restants, sans compter la fonction en cours (mettre = a N-1 au dpart)
//Spcifier LastValeur a 0 pour le dpart. Correspond en fait a la valeur prcdente, c'est a partir de lui qu'il faut chercher
//Indexs : correspond a la somme de tous les index prcdents    (TCandidats pour simplifier les calculs)

       //TabValeurs : contient (a la fin) les valeurs correspondants au NiemeCach dans l'ordre dcroissant
//pour trouver si une valeur est NiemeCach, on fait la somme des index prcdents. si le nombre d'indexs
//de cette somme est plus grand que N, la valeurs ne fait pas partie du NiemeCach des index prcdents

var Valeur : TSmallNum;
    IndexsNouv : TCandidats;
begin
  Result:=False;
  Valeur:=LastValeur;                     //Commencer la recherche a partir de la valeur prcdente +1 ( attention : LastValeur = [1..9] mais Valeur = [0..8]
  while (Valeur<9-NRestant) and not Result do
  begin

    if (Occurence[Zone,Valeur+1,NoElement]<=N) and (Occurence[Zone,Valeur+1,NoElement]>1) then    //Verifie rapidement si la valeur peut convenir
    begin
      IndexsNouv:=Indexs+GetIndexs(Valeur+1, NoElement, Zone);    //En faisant la somme des set of Index, on vrifie si les indexs sont identiques, ou si un index est inclus dans l'autre.
      if CompteCandidats(IndexsNouv)<=N then
      begin
        TabValeurs[NRestant]:=Valeur+1;
        if NRestant=0 then
        begin
          Result:=SupprimerSaufValeursInIndexs(TabValeurs,IndexsNouv,NoElement,Zone);
          if Result and AffMess then MessageNiemeCache(TabValeurs,IndexsNouv,N,NoElement,Zone);
        end
        else
          Result:=TrouverNiemeCache(TabValeurs,N,NRestant-1,Valeur+1,IndexsNouv,NoElement,Zone);
      end;
    end;

    inc(Valeur);
  end;
end;

---------------------------------------------------------------------------------------------}



function TesterCandidatsJusquaNieme (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum):TResultat;
var NoElement, NoElemTemp, OrdZone, NbVides, NCacheMax, NCandMax : TSmallNum;
begin
  Result.NoElements:=0;
  Result.Index:=0;

  for NoElement:=1 to TAILLE_SUDOKU do
  begin
    NoElemTemp:=((NoElement+LastElement-2) mod TAILLE_SUDOKU)+1;
    for OrdZone:=0 to 2 do
    begin
      NbVides:=CompterCasesNonTrouveInZone(Grille[TZone(OrdZone),NoElemTemp]);
      NCacheMax:=(NbVides-1) div 2;
      NCandMax:=NbVides-NCacheMax-1;
      if ((NCandMax>=2) and TrouverCandidatJusquaNieme(Grille[TZone(OrdZone),NoElemTemp],Occur,NoElemTemp,TZone(OrdZone),NCandMax,1,1,0,0)) or
       ((NCacheMax>=2) and TrouverCacheJusquaNieme(Grille[TZone(OrdZone),NoElemTemp],Occur,NoElemTemp,TZone(OrdZone),NCacheMax,1,1,0,0)) then
      begin
        LastElement:=NoElemTemp;
        Result.NoElements:=1 shl (NoElemTemp-1);
        Result.Zone:=TZone(OrdZone);
        Exit;
      end;
    end;
  end;
end;





function PositionDesIndexs (Indexs : TEnsemble): TSmallNum;
//Retourne la "position" des indexs : si les indexs sont groups dans un bloc de 3 ([1,2,3] ou [4,5,6]),
//renvoie 1 ou 2 ou 3. Sinon renvoie 0
//Exemples :
//      [1,2] > 1    [4,6] > 2     [7,8,9] > 3      [1,2,5] > 0      [4,8] > 0
//var x : TSmallNum;
begin
  Result:=TAILLE_BLOC;
  while (Result>0) and ((Indexs and (POS_BLOC shl (TAILLE_BLOC*(Result-1))))<>Indexs) do
    dec(Result);

 {ANCIEN
  for x:=0 to TAILLE_BLOC-1 do
    if Indexs<=[(x*TAILLE_BLOC)+1..(x*TAILLE_BLOC)+TAILLE_BLOC] then    //si indexs est dans [1,2,3] ou [4,5,6] ...
    begin
      Result:=x+1;
      Break;
    end;     }
end;





function PositIndexDansBloc (Indexs : TEnsemble): TSmallNum;
//Retourne un chiffre permettant de trouver la colonne dans laquelle les indexs sont. (si ils sont disposs en colonne !)
// cette fonction dpends de la constante POS_BLOC2
begin
  Result:=TAILLE_BLOC;
  while (Result>0) and ((Indexs and (POS_BLOC2 shl ((Result-1))))<>Indexs) do
    dec(Result);

 { Result:=0;
  if Indexs<=[1,4,7] then Result:=1
  else
    if Indexs<=[2,5,8] then Result:=2
    else
      if Indexs<=[3,6,9] then Result:=3; }
end;






function ConvertirIndexDepartToIndexArrivee(const IndexsDep : TEnsemble ;const NoElementDep : TSmallNum;const ZoneDepart : TZone;
                                              out IndexsArr : TEnsemble ; out  NoElementArr : TSmallNum; out  ZoneArrivee: TZone) : Boolean;
var Posit, Index, IndexTemp : TSmallNum;
begin
  Result:=False;
  Posit:=PositionDesIndexs(IndexsDep);
  case ZoneDepart of
    zsLigne : if Posit<>0 then               //Disposs dans un seul bloc
              begin
                Result:=True;
                ZoneArrivee:=zsBloc;
                NoElementArr:=((NoElementDep-1) div TAILLE_BLOC)*TAILLE_BLOC + Posit;
                IndexTemp:=((NoElementDep-1) mod TAILLE_BLOC)*TAILLE_BLOC;
                IndexsArr:=0;
                for Index:=0 to TAILLE_SUDOKU-1 do
                if IsBitSet(IndexsDep,Index) then
                  SetBit(IndexsArr,(Index mod TAILLE_BLOC)+IndexTemp);
              end;
    zsColonne : if Posit<>0 then               //Disposs dans un seul bloc
                begin
                  Result:=True;
                  ZoneArrivee:=zsBloc;
                  NoElementArr:=(NoElementDep-1) div TAILLE_BLOC + (Posit-1)*TAILLE_BLOC +1;
                  IndexTemp:=(NoElementDep-1) mod TAILLE_BLOC;
                  IndexsArr:=0;
                  for Index:=0 to TAILLE_SUDOKU-1 do
                  if IsBitSet(IndexsDep,Index) then
                    SetBit(IndexsArr,(Index mod TAILLE_BLOC)*TAILLE_BLOC+IndexTemp);
                end;
    zsBloc :  if Posit<>0 then               //Disposes en ligne
              begin
                Result:=True;
                ZoneArrivee:=zsLigne;
                NoElementArr:=Posit + ((NoElementDep-1) div TAILLE_BLOC)*TAILLE_BLOC;
                IndexTemp:=((NoElementDep-1) mod TAILLE_BLOC)*TAILLE_BLOC;
                IndexsArr:=0;
                for Index:=0 to TAILLE_SUDOKU-1 do
                if IsBitSet(IndexsDep,Index) then
                  SetBit(IndexsArr,(Index mod TAILLE_BLOC)+IndexTemp);
              end
              else
              begin
                Posit:=PositIndexDansBloc(IndexsDep);
                if Posit<>0 then             //Disposes en colonne
                begin
                  Result:=True;
                  ZoneArrivee:=zsColonne;
                  NoElementArr:=Posit + ((NoElementDep-1) mod TAILLE_BLOC)*TAILLE_BLOC;
                  IndexTemp:=((NoElementDep-1) div TAILLE_BLOC)*TAILLE_BLOC;
                  IndexsArr:=0;
                  for Index:=0 to TAILLE_SUDOKU-1 do
                  if IsBitSet(IndexsDep,Index) then
                    SetBit(IndexsArr,Index div TAILLE_BLOC + IndexTemp);
                end;
              end;
    end;
end;






procedure MessageInteraction(Valeur, NoElementDep : TSmallNum; ZoneDepart : TZone; NoElementArr : TSmallNum; ZoneArrivee : TZone);
begin
  if EtapesResolution <> '' then
  EtapesResolution:=EtapesResolution+#13#10;

  EtapesResolution:=EtapesResolution+' - Suppression de candidats dans'+ZoneToStr[ZoneArrivee]+IntToStr(NoElementArr)+' car le '+IntToStr(Valeur)
                                           +' de cette zone est dans'+ZoneToStr[ZoneDepart]+IntToStr(NoElementDep);
end;






function Interaction(var Grille : PZonesSudoku; var Occur : TOccurence; NoElementDep : TSmallNum; ZoneDepart : TZone) : TResultat;
//cherche dans une ligne, une colonne ou un bloc (suivant Zone) NoElement si les valeurs sont disposes
//de maniere a interagir avec une ligne, un bloc ou une colonne.
//pour une ligne, si les valeurs sont disposes comme 123 ou 456 ou 789, enleve les candidats du bloc.
//pour une colonne, si les valeurs sont disposes comme 123 ou 456 ou 789, enleve les candidats du bloc.
//pour un bloc, si ils sont disposs en 123 ou 456 ou 789, enleve les candidats sur la ligne
//              si ils sont disposs en 147 ou 258 ou 369, enleve les candidats sur la colonne
//renvoie un Resultat si il y a eu modification.

var Valeur, NoElementArr : TSmallNum;
    ZoneArrivee : TZone;
    Test : Boolean;
    IndexsArr : TEnsemble;
begin
  Result.Index:=0;
  Result.NoElements:=0;

  for Valeur:=1 to TAILLE_SUDOKU do
    if (Occur[ZoneDepart,NoElementDep,Valeur] in [2..TAILLE_BLOC]) and ConvertirIndexDepartToIndexArrivee(GetIndexs(Grille[ZoneDepart,NoElementDep],Valeur),NoElementDep,ZoneDepart,IndexsArr,NoElementArr,ZoneArrivee) then
    begin
      if Occur[ZoneArrivee,NoElementArr,Valeur]>Occur[ZoneDepart,NoElementDep,Valeur] then
      begin
        Test:=SupprimerDansIndexsInZone(Grille[ZoneArrivee,NoElementArr],Occur,1 shl (Valeur-1),ENSEMBLE_VALEURS-IndexsArr, NoElementArr, ZoneArrivee);
        if Test then
        begin
          Result.NoElements:=1 shl (NoElementArr-1);
          Result.Zone:=ZoneArrivee;
          if AfficherEtapesResolution then MessageInteraction(Valeur,NoElementDep, ZoneDepart, NoElementArr, ZoneArrivee );
          Exit;
        end;
      end;
    end;
end;




function TesterInteraction (var Grille : PZonesSudoku; var Occur : TOccurence; var LastElement : TSmallNum) : TResultat;
//argument = [0..8]
//regarde si c'est possible de simplifier une ligne ou une colonne grace a un bloc :
//    125   8   9
//    235  23   4
//    12    6   7
//     9    x   x
//     7    x   x
//    56    x   x
//    53    x   x               Simplifiera le 56 et 53 en 6 et 3 car ds le bloc en haut, le 5 est forcment sur la colonne 1
//Ou le contraire : regarde si on peut simplifier un bloc grace a une ligne :
//    13    8   9
//    23   53   43
//    12    6   7
//     4    x   x
//    25    x   x
//    56    x   x
//    54    x   x        //Simplifiera le 53 et 43 du bloc en haut car dans la colonne de gauche, le 3 est forcment dans le bloc en haut


var NoElem, OrdZone, NoElemTemp : TSmallNum;
begin
  Result.NoElements:=0;
  Result.Index:=0;

  for NoElem:=1 to TAILLE_SUDOKU do
  begin
    NoElemTemp:=((NoElem+LastElement-2) mod TAILLE_SUDOKU)+1;
    for OrdZone:=0 to 2 do
    begin
      Result:=Interaction(Grille,Occur,NoElemTemp,TZone(OrdZone));
      if Result.NoElements<>0 then
      begin
        LastElement:=NoElemTemp;
        Exit;
      end;
    end;
  end;

end;




type                  // ce type est utile seulement pour les X-Wing et compagnie

  TElementIndexs = Record
     NoElement : TSmallNum;
     Indexs    : TEnsemble;
  end;





function VerifierSiLiaisonsOK (Elements : array of TElementIndexs; N : TSmallNum) : boolean ;
var NoIndex, NoElem, NbBits : TSmallNum;
begin
  for NoIndex:=0 to TAILLE_SUDOKU-1 do
  begin
    NbBits:=0;

    for NoElem:=0 to N-1 do
      inc(NbBits,((Elements[NoElem].Indexs shr NoIndex) and 1));

    Result:=(NbBits<>1);
    if not(Result) then Break;
  end;
end;





function OppZone ( Zone : TZone) : TZone;
begin
  if Zone=zsLigne then Result:=zsColonne
                  else Result:=zsLigne;
end;




procedure MessageEtoileNBranche (N, Valeur : TSmallNum; Zone : TZone; SommeElems, SommeIndexs : TEnsemble);
//Message pour l'toile a n branche
//zone doit etre la zone d'arrive, dans laquelle les candidats sont supprims
const TypeToStr : array [2..5] of string = (' X-Wing ', ' SwordFish ', ' JellyFish ', ' SANS NOM ');

var Elements, Indexs : string;
    X : TSmallNum;
begin
  Elements:='';

  for X:=0 to TAILLE_SUDOKU do
    if IsBitSet(SommeElems, X) then Elements:=Elements+', '+IntToStr(X+1);
  Delete(Elements,1,2);

  Indexs:='';
  for X:=0 to TAILLE_SUDOKU do
    if IsBitSet(SommeIndexs, X) then Indexs:=Indexs+', '+IntToStr(X+1);
  Delete(Indexs,1,2);

  if EtapesResolution <> '' then
  EtapesResolution:=EtapesResolution+#13#10;

  EtapesResolution:=EtapesResolution+' - Suppression de candidats dans'+ZonePluToStr[Zone]+Elements+
                                     ' car'+TypeToStr[N]+'de '+IntToStr(Valeur)+' dans'+ZonePluToStr[OppZone(Zone)]+Indexs;
end;




function SupprimerCandEtoile (var Grille : PZonesSudoku; var Occ : TOccurence; Valeur, N : TSmallNum; Elems : array of TElementIndexs; SommeIndexs : TEnsemble; Zone : TZone) : boolean;
var X : TSmallNum;        //variable pour les boucles
    NoElems : TEnsemble;
    Modif : Boolean;      // variable pour eviter le probleme des booleans (result or test : test ne sera pas effectu si result = true)
begin
  Result:=False;
  NoElems:=0;

  for X:=0 to N-1 do
    NoElems:=NoElems or (1 shl (Elems[X].NoElement-1));

  for X:=0 to TAILLE_SUDOKU-1 do
    if IsBitSet(SommeIndexs, X) then
    begin
      Modif:=SupprimerDansIndexsInZone(Grille[Zone,X+1],Occ,1 shl (Valeur-1),ENSEMBLE_VALEURS - NoElems,X+1,Zone);
      Result:=Result or Modif;
    end;

  if AfficherEtapesResolution and Result then MessageEtoileNBranche(N,Valeur,Zone,SommeIndexs, NoElems);
end;





function SommeElemsTableau (Tab : Array of TElementIndexs; N : TSmallNum):TEnsemble;
var X : TSmallNum;
begin
  Result:=0;
  For X:=0 to N-1 do
    SetBit(Result,Tab[X].NoElement-1);
end;





function  EtoileANBranches(var Grille : PZonesSudoku; var Occ : TOccurence; NoElemPrec, Valeur : TSmallNum; TypeZone : TZone;
                            NMax, N : TSmallNum; var Elems : array of TElementIndexs; SommeIndexs : TEnsemble; out SommeElems : TEnsemble) : Boolean;
//Trouve des X-Wing, SwordFish, JellyFish de 'Valeur' dans chaque zone de type TypeZone :
//peut trouver un XWing de 3 disposs en ligne : Valeur=3, Res.Zone=zsLigne
//Recursif
//Initialiser :
//  NoElemPrec a 0
//  NMax a la valeur max de branche (pour une grille 9*9, 4 suffit ((9-1) div 2)
//  N a 1
//  Elems doit etre vide et de taille = (ou >) a NMax
//  SommeIndexs a 0
//Renvoie True en cas de succes (une etoile a t trouve, et des candidats supprims ...)
//SommeElems permet de retourner les elements


var NoElemInter, NbIndexs : TSmallNum;
    IndexsInter, SommeIndexsModif : TEnsemble;
begin
  Result:=False;

  for NoElemInter:=NoElemPrec+1 to TAILLE_SUDOKU do
    if (Occ[TypeZone,NoElemInter,Valeur]<NMax) and (Occ[TypeZone,NoElemInter,Valeur]>1) then
    begin
      IndexsInter:=GetIndexs(Grille[TypeZone,NoElemInter],Valeur);
      SommeIndexsModif:=SommeIndexs or IndexsInter;
      NbIndexs :=CompteCandidats(SommeIndexsModif);

      if NbIndexs<=NMax then
      begin
        Elems[N-1].NoElement:=NoElemInter;
        Elems[N-1].Indexs:=IndexsInter;

        if N=NbIndexs then
        begin
          if VerifierSiLiaisonsOK(Elems,N) then
          begin
            Result:=SupprimerCandEtoile(Grille,Occ,Valeur,N,Elems,SommeIndexsModif,OppZone(TypeZone));
            if Result then //Si on a trouv, on enregistre les elements
              SommeElems:=SommeElemsTableau(Elems,N);
          end;


        end
        else
          if N<NMax then Result:=EtoileANBranches(Grille,Occ,NoElemInter,Valeur,TypeZone,NMax, N+1, Elems, SommeIndexsModif,SommeElems);

        if Result then Break;
      end;
    end;
end;







function TesterEtoileANBranche (var Grille : PZonesSudoku; var Occ : TOccurence; Var LastValeur : TSmallNum): TResultat;
//Teste si il existe une toile a 2 branches (X-Wing), 3 branches (SwordFish), 4 branches (JellyFish) dans toute la grille
var Valeur, ValTemp : TSmallNum;
    Elems : array [0..TAILLE_SUDOKU div 2] of TElementIndexs;
    Trouve: boolean;
begin
  Result.Index:=0;
  Result.NoElements:=0;
  Result.Zone:=zsBloc;            //impossible

  for Valeur:=1 to TAILLE_SUDOKU do
  begin
    ValTemp:=((Valeur+LastValeur-2) mod TAILLE_SUDOKU)+1;
    Trouve:=EtoileANBranches(Grille,Occ,0,Valeur,zsLigne,TAILLE_SUDOKU div 2,1,Elems,0, Result.NoElements);
    if Trouve then Result.Zone:=zsLigne;
    Trouve:=Trouve or EtoileANBranches(Grille,Occ,0,Valeur,zsColonne,TAILLE_SUDOKU div 2,1,Elems,0, Result.NoElements);

    if Trouve then
    begin
      if Result.Zone=zsBloc then Result.Zone:=zsColonne;
      LastValeur:=ValTemp;
      Break;
    end;
  end;
end;





///////////////////////////////////////////////////////////////////////////////////////////
//                                   Brute forces (BF)                                   //
//                                                                                       //
// Voici plusieurs solveurs bruteforce. Je les ai gard parce qu'ils ne fonctionnent pas //
// de la meme maniere.                                                                   //
///////////////////////////////////////////////////////////////////////////////////////////



{------------------------------------------------------------}
{                                                            }
{             Methodes de rsolution Brute Force 1           }
{                                                            }
{------------------------------------------------------------}

//Ce solveur BF est bas sur une rcursivit. une variable varie de 1  81 et passe toutes
//les cases. La grille est donne en parametre SANS ETRE de type var. A chaque case non
//trouve, dans la fonction N, on passe une par une les valeurs, et on rappelle la fonction N+1
// avec cette valeur. La fonction N+1 ajoute la valeur, et continue la rcursivit.
//Si N+1 est infructueuse, elle se termine, et N passe a la valeur suivante. Comme le
//parametre n'est pas de type var, N ne contient pas la valeur prcdemment ajoute en N+1
//Pour connaitre les candidats d'une case, il suffit de regarder les valeurs prises par
//la case. Celle ci est maintenue a jour grace a la procedure AjouterValeur, qui supprime
//les candidats.






function EnleverCandidatBF(var CaseCand : TEnsemble; Valeur : TSmallNum) : Boolean;
//Enleve le candidat 'Valeur' de la case 'Ligne, Colonne' de CaseCand
//Renvoie True en cas d'erreur (si il y a suppression d'un candidat unique)

begin
  ClearBit(CaseCand,Valeur-1);
  Result:=CaseCand=0;
end;





function AjouterValeurBF (var Grille : TGrilleSudoku; Valeur, Ligne, Colonne : TSmallNum):Boolean;
//Procedure qui ajoute une valeur 'Valeur' en position 'Ligne, Colonne'
//Elle ajoute la valeur dans Grille, enleve les candidats appropris
//Dans ce solveur, les valeurs sont places de gauche a droite et de haut en bas. Donc
//il est inutile de supprimer les candidats avant la valeur en cours.
//Les candidats supprims sont ceux jusqu'a la fin de la ligne, jusqu'a la fin de la colonne, et jusqu'a la fin du bloc.

var NoLigne, NoColonne : TSmallNum;
begin
  Result:=False;

  for NoColonne:=Colonne+1 to TAILLE_SUDOKU do
    if not(Grille[Ligne,NoColonne].Trouve) then
      Result:=Result or EnleverCandidatBF(Grille[Ligne,NoColonne].Cand,Valeur);

  //Supprime les candidats dans la meme ligne

  if Not(Result) then
  for NoLigne:=((((Ligne-1) div TAILLE_BLOC)+1)*TAILLE_BLOC + 1) to TAILLE_SUDOKU do
    if not(Grille[NoLigne,Colonne].Trouve) then
      Result:=Result or EnleverCandidatBF(Grille[NoLigne,Colonne].Cand,Valeur);

  //Supprime les candidats dans le meme colonne

  if Not(Result) then
  for NoLigne:=Ligne+1 to (((Ligne-1) div TAILLE_BLOC + 1)*TAILLE_BLOC) do
  for NoColonne:=(((Colonne-1) div TAILLE_BLOC)*TAILLE_BLOC + 1) to ((((Colonne-1) div TAILLE_BLOC)+1)*TAILLE_BLOC) do
    if not(Grille[NoLigne,NoColonne].Trouve) then
      Result:=Result or EnleverCandidatBF(Grille[NoLigne,NoColonne].Cand,Valeur);

  //Supprime les candidats dans le meme bloc

  //On N'a pas besoin de mettre la valeur dans la case parce qu'elle ne nous sert plus
end;






procedure ConvertToPos(const NoCase : Byte; out Ligne,Colonne : TSmallNum);
begin
  Ligne:=((NoCase-1) div TAILLE_SUDOKU)+1;
  Colonne:=((Nocase-1) mod TAILLE_SUDOKU)+1
end;





function ResoudreBruteForce (Sud : TGrilleSudoku; NoCase : Byte; LastValeur : TSmallNum; var GrilleSortie : TGrilleSudoku):boolean;
var Ligne, Colonne, Valeur : TSmallNum;
    Erreur : Boolean;
begin
  Result:=False;
  Erreur:=False;

  ConvertToPos(NoCase,Ligne,Colonne);

  if LastValeur<>0 then
    if Colonne>1 then
      Erreur:=AjouterValeurBF(Sud,LastValeur,Ligne,Colonne-1)
    else
      Erreur:=AjouterValeurBF(Sud,LastValeur,Ligne-1,TAILLE_SUDOKU);


  if Not(Erreur) then
  if (NoCase >= TAILLE_SUDOKU*TAILLE_SUDOKU) then
  begin
    GrilleSortie[Ligne,Colonne].Cand:=Sud[Ligne,Colonne].Cand;
    Result:=True;
           //si on est a la case 81, on est sur qu'il y a un candidat dans la case,
           //sinon il y aurait eu une erreur lors de la suppression du dernier candidat
  end
  else
  begin
    if Sud[Ligne,Colonne].Trouve then
      Result:=ResoudreBruteForce(Sud,NoCase+1,0,GrilleSortie)
    else
    begin
      for Valeur:=0 to TAILLE_SUDOKU-1 do
        if IsBitSet(Sud[Ligne,Colonne].Cand,Valeur) then
        begin
          Result:=ResoudreBruteForce(Sud,NoCase+1,Valeur+1,GrilleSortie);

          if Result then
          begin
            GrilleSortie[Ligne,Colonne].Cand:=1 shl Valeur;
            Break;
          end;
        end;
    end;
  end;

end;



{------------------------------------------------------------}
{                                                            }
{             Methodes de rsolution Brute Force 2           }
{                                                            }
{------------------------------------------------------------}
//Cette mthode parcours les cases de la meme maniere que le BF1.
//La diffrence est que pour trouver les candidtas d'une case, on les calcule a chaque fois
//(l'ide de calculer pour chaque case est l'ide de FullSpeed sur le forum)
//Donc, pour chaque ajout de valeur, on n'a plus besoin de supprmier les candidats.
//Cette mthode est plus lente.



function TrouverCandidatsCase(var Grille : TGrilleSudoku; Ligne, Colonne : TSmallNum):TEnsemble;
var X : TSmallNum;
    CaseSud : TCaseSudoku;
begin
  Result:=ENSEMBLE_VALEURS;
  for X:=1 to TAILLE_SUDOKU do
  begin
    if Grille[Ligne, X].Trouve then Result:=Result and (not(Grille[Ligne, X].Cand));
    if Grille[X, Colonne].Trouve then Result:=Result and (not(Grille[X, Colonne].Cand));
    CaseSud:=Grille[((((Ligne-1) div TAILLE_BLOC)*TAILLE_BLOC)+(X-1) div TAILLE_BLOC + 1),((((Colonne-1) div TAILLE_BLOC)*TAILLE_BLOC)+(X-1) mod TAILLE_BLOC + 1)];
    if CaseSud.Trouve then Result:=Result and (not(CaseSud.Cand))
  end;
end;




function ResoudreBruteForce2 (var Sud : TGrilleSudoku; NoCase : Byte):boolean;
var Ligne, Colonne, Valeur : TSmallNum;
    Erreur : Boolean;
    CandCase : TEnsemble;
begin
  Result:=False;
  Erreur:=False;

  ConvertToPos(NoCase,Ligne,Colonne);

  if Not(Erreur) then
  if not(Sud[Ligne,Colonne].Trouve) then
  begin
    CandCase:=TrouverCandidatsCase(Sud,Ligne,Colonne);
   // showmessage(IntTostr(CandCase)+', l : '+IntTostr(Ligne)+', c : '+IntTostr(Colonne));
    if CandCase>0 then
    begin
      Sud[Ligne,Colonne].Trouve:=True;

      for Valeur:=0 to TAILLE_SUDOKU-1 do
        if IsBitSet(CandCase,Valeur) then
        begin
          Sud[Ligne,Colonne].Cand:=1 shl Valeur;

          if NoCase <TAILLE_SUDOKU*TAILLE_SUDOKU then
            Result:=ResoudreBruteForce2(Sud,NoCase+1)
          else
            Result:=True;

          if Result then Break;
        end;

      Sud[Ligne,Colonne].Trouve:=Result;
    end;
  end
  else
  begin
    if NoCase <TAILLE_SUDOKU*TAILLE_SUDOKU then
      Result:=ResoudreBruteForce2(Sud,NoCase+1)
    else
      Result:=True;
  end;


end;



{------------------------------------------------------------}
{                                                            }
{             Methodes de rsolution Brute Force 3           }
{                                                            }
{------------------------------------------------------------}

//Le BF 3 est similaire au BF1, sauf que lors de l'ajout d'une valeur, on verifie si une autre
//valeur peut etre place, et si oui, on la place, ainsi de suite.
//L'avantage est que plusieurs cases perdront des candidats, mais par contre, cela oblige
//a chaque placement de valeurs, a verifier la ligne entiere, la colonne entiere,
//et le bloc entier.
//Le deuxieme changement est que la grille est celle contenant les pointeurs, donc en cas d'ajout
//de valeurs, il faut aussi pouvoir restaurer la grille d'avant cet ajout.
//Il y a donc des procedures d'enregistrement et de restauration de grille.

//La raison principale de ce changement est que j'ai remarqu que pour certaines grilles, l'ajout
//de 4 valeurs seulement pouvait permettre la rsolution.

//Cette mthode de rsolution est beaucoup plus rapide que la premiere

//Temps :    BF2 > BF1 >> BF3




function EnleverCandidatBF3(var Grille : PZonesSudoku; Valeur, Index, NoElement : TSmallNum; Zone : TZone; Id : Byte) : Boolean;
//Enleve le candidat 'Valeur' de la case 'Ligne, Colonne' de CaseCand
//Renvoie True en cas d'erreur (si il y a suppression d'un candidat unique)

var V : TSmallNum;
begin
  Result:=False;
  if not(Grille[Zone,NoElement,Index]^.Trouve) then
  begin
    ClearBit(Grille[Zone,NoElement,Index]^.Cand,Valeur-1);
    Result := Grille[Zone,NoElement,Index]^.Cand = 0;

    if not(Result) then
      for V:=0 to TAILLE_SUDOKU-1 do
        if Grille[Zone,NoElement,Index]^.Cand=(1 shl V) then
        begin
          Result:=AjouterValeurBF3(Grille,V+1,Index,NoElement,Zone);
          Break;
        end;
  end;
end;






function AjouterValeurBF3 (var Grille : PZonesSudoku; Valeur, Index, NoElement : TSmallNum; Zone : TZone):Boolean;
//Procedure qui ajoute une valeur 'Valeur' en position 'Ligne, Colonne'
//Elle ajoute la valeur dans Grille, enleve les candidats appropris
//Les candidats supprims sont ceux jusqu'a la fin de la ligne, jusqu'a la fin de la colonne, et jusqu'a la fin du bloc.
//Les valeurs du BF sont places dans l'ordre d'en haut a gauche jusqu'en bas a droite.

var X, NoLigne, NoColonne, NoBloc : TSmallNum;
    CandTmp : TEnsemble;
    Id : Byte;
begin
  Grille[Zone,NoElement,Index]^.Trouve:=True;
  CandTmp:= 1 shl (Valeur-1);
  Grille[Zone,NoElement,Index]^.Cand:=CandTmp;

  IndexToPosition(Index, NoElement, Zone, NoLigne, NoColonne);
  NoBloc:=LCToBloc(NoLigne,NoColonne);

  Id:=(NoLigne-1)*TAILLE_SUDOKU+NoColonne;

  Result:=False;


  for X:=1 to TAILLE_SUDOKU do
  begin

    //Pour enlever un candidat, il faut que Result = false, que la case en cours ne soit pas
    //la case initiale, et que les candidats de la case en cours ne soit pas [Valeur]
    //(sinon il y aura erreur) qu'il y ait "Trouve" ou pas

    Result:=Result or (Grille[Zone,NoElement,Index]<>Grille[zsLigne,NoLigne,X]) and
            ((Grille[zsLigne,NoLigne,X]^.Cand=CandTmp)
             or EnleverCandidatBF3(Grille,Valeur,X,NoLigne,zsLigne,Id));
    Result:=Result or (Grille[Zone,NoElement,Index]<>Grille[zsColonne,NoColonne,X]) and
            ((Grille[zsColonne,NoColonne,X]^.Cand=CandTmp)
             or EnleverCandidatBF3(Grille,Valeur,X,NoColonne,zsColonne,Id));
    Result:=Result or (Grille[Zone,NoElement,Index]<>Grille[zsBloc,NoBloc,X]) and
            ((Grille[zsBloc,NoBloc,X]^.Cand=CandTmp)
             or EnleverCandidatBF3(Grille,Valeur,X,NoBloc,zsBloc,Id));

    if Result then Break;

    //Result vaut true si il y a une erreur (la supression d'un candidat seul rendant la case vide)
    //si result vaut true, tout s'arrete.
  end;
end;





procedure SauverPointeurs(var Pointeurs : PZonesSudoku;var Grille : TGrilleSudoku);
var i,j : TSmallNum;
begin
  for i:=TAILLE_SUDOKU downto 1 do
  for j:=TAILLE_SUDOKU downto 1 do
    Grille[i,j]:=Pointeurs[zsLigne,i,j]^;
end;




procedure ChargerGrille(var Grille : TGrilleSudoku; var Pointeurs : PZonesSudoku);
var i,j : TSmallNum;
begin
  for i:=TAILLE_SUDOKU downto 1 do
  for j:=TAILLE_SUDOKU downto 1 do
    Pointeurs[zsLigne,i,j]^:=Grille[i,j];
end;




function ResoudreBruteForce3 (var Ptrs : PZonesSudoku; NoCase : Byte):boolean;
var Ligne, Colonne, Valeur : TSmallNum;
    Erreur : Boolean;
    GSecours : TGrilleSudoku;
begin
  Result:=False;

  ConvertToPos(NoCase,Ligne,Colonne);


  if (NoCase >= TAILLE_SUDOKU*TAILLE_SUDOKU) then
  begin
    Result:=True;
           //si on est a la case 81, on est sur qu'il y a un candidat dans la case,
           //sinon il y aurait eu une erreur lors de la suppression du dernier candidat
  end
  else
  begin
    if Ptrs[zsLigne,Ligne,Colonne]^.Trouve then
      Result:=ResoudreBruteForce3(Ptrs,NoCase+1)
    else
    begin
      SauverPointeurs(Ptrs,GSecours);

      for Valeur:=1 to TAILLE_SUDOKU do
        if IsBitSet(Ptrs[zsLigne,Ligne,Colonne]^.Cand,Valeur-1) then
        begin
          Erreur:=AjouterValeurBF3(Ptrs,Valeur,Colonne,Ligne,zsLigne);
          if Not(Erreur) then
          begin
            Result:=ResoudreBruteForce3(Ptrs,NoCase+1);
            if Result then Break;
          end;
          ChargerGrille(GSecours,Ptrs);
        end;

    end;
  end;

end;





end.
